@@ -2,6 +2,7 @@
 Test UpdateCapabilities.
 """
 
+from functools import partial
 import dbus
 
 from twisted.words.xish import xpath, domish
@@ -247,14 +248,20 @@ def run_mixed_test (q, bus, conn, stream):
     check_caps(namespaces, JINGLE_CAPS)
 
 if __name__ == '__main__':
-    exec_test(lambda q, b, c, s:
-        run_test (q, b, c, s,
-            cs.CHANNEL_TYPE_STREAMED_MEDIA, cs.CHANNEL_IFACE_MEDIA_SIGNALLING,
-            cs.INITIAL_AUDIO, cs.INITIAL_VIDEO), do_connect=False)
+    exec_test(
+        partial(run_test,
+            media_channel_type=cs.CHANNEL_TYPE_STREAMED_MEDIA,
+            media_interface=cs.CHANNEL_IFACE_MEDIA_SIGNALLING,
+            initial_audio=cs.INITIAL_AUDIO,
+            initial_video=cs.INITIAL_VIDEO),
+        do_connect=False)
 
-    exec_test(lambda q, b, c, s:
-        run_test (q, b, c, s,
-            cs.CHANNEL_TYPE_CALL, cs.CHANNEL_TYPE_CALL,
-            cs.CALL_INITIAL_AUDIO, cs.CALL_INITIAL_VIDEO), do_connect=False)
+    exec_test(
+        partial(run_test,
+            media_channel_type=cs.CHANNEL_TYPE_CALL,
+            media_interface=cs.CHANNEL_TYPE_CALL,
+            initial_audio=cs.CALL_INITIAL_AUDIO,
+            initial_video=cs.CALL_INITIAL_VIDEO),
+        do_connect=False)
 
     exec_test(run_mixed_test, do_connect=False)
